////////////////////////////////////////////////////////////////////////////////////////////////////////
//  July 2015
//  Author: Juan Jose Chong <juan.chong@analog.com>
//  Modified by: Martin . Christensen / Egil Gundersen
////////////////////////////////////////////////////////////////////////////////////////////////////////
//  ADIS16448.cpp
////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
//  This library provides all the functions necessary to interface the ADIS16448 IMU with an 
//  8-Bit Atmel-based Arduino development board. Functions for SPI configuration, reads and writes,
//  and scaling are included. This library may be used for the entire ADIS164XX family of devices 
//  with some modification.
//
//  This example is free software. You can redistribute it and/or modify it
//  under the terms of the GNU Lesser Public License as published by the Free Software
//  Foundation, either version 3 of the License, or any later version.
//
//  This example is distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
//  FOR A PARTICULAR PURPOSE.  See the GNU Lesser Public License for more details.
//
//  You should have received a copy of the GNU Lesser Public License along with 
//  this example.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "ADIS16448.h"

////////////////////////////////////////////////////////////////////////////
// Constructor with configurable CS, DR, and RST
////////////////////////////////////////////////////////////////////////////
// CS - Chip select pin
// DR - DR output pin for data ready
// RST - Hardware reset pin
////////////////////////////////////////////////////////////////////////////
ADIS16448::ADIS16448(int CS, int DR, int RST) {
	_CS = CS;
	_DR = DR;
	_RST = RST;
}

void ADIS16448::init(SemaphoreHandle_t SPImut) {
	_gyroRange = GyroRange::PM1000;
	_SPImut = SPImut;
	uint8_t filterSizeB = 5;
	uint8_t decimationRate = 2;
	uint8_t useInternalSamplingClock = 1;

	//Set default pin states
	pinMode(_CS, OUTPUT);		// Set CS pin to be an output
	pinMode(_DR, INPUT);		// Set DR pin to be an input
	pinMode(_RST, OUTPUT);		// Set RST pin to be an output
	digitalWrite(_CS, HIGH);	// Initialize CS pin to be high
	digitalWrite(_RST, HIGH);	// Initialize RST pin to be high

	// Initialize SPI bus
	_ADIS_SPIsettings = SPISettings(ADIS_SPI_FREQ, MSBFIRST, SPI_MODE3);
	SPI.begin();
	delay(100);						// Give the part time to start up
	this->regWrite(MSC_CTRL, 0x6);	// Enable Data Ready on IMU
	delay(20);
	this->regWrite(SENS_AVG, (((uint8_t)_gyroRange << 8) | filterSizeB)); // Set gyro scale and digital filter on IMU
	delay(20);
	this->regWrite(SMPL_PRD, (decimationRate << 8 | useInternalSamplingClock)); // Set Decimation on IMU
	delay(20);
}


float ADIS16448::calcGravity(char axis, int8_t sign)
{
	int samples = 0;
	float gravity = 0.0;
	int startupSamples = 50;	// Number of samples to average
	while (digitalRead(_DR));	// Wait for new data (not ready = data update)
	while (samples < startupSamples) {
		if (digitalRead(_DR)) {
			float currentAcc = this->readAccel(axis, sign);
			samples++;
			gravity += currentAcc;
			delay(20);
			while (digitalRead(_DR)); // Wait for new data (not ready = data update)
		}
	}
	return gravity / startupSamples;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Reads two bytes (one word) in two sequential registers over SPI
////////////////////////////////////////////////////////////////////////////////////////////
// regAddr - address of register to be read
// return - (int) signed 16 bit 2's complement number
////////////////////////////////////////////////////////////////////////////////////////////
int16_t ADIS16448::regRead(uint8_t regAddr) {
	//Read registers using SPI
	xSemaphoreTake(_SPImut, portMAX_DELAY);
	SPI.beginTransaction(_ADIS_SPIsettings);

	// Write register address to be read
	digitalWrite(_CS, LOW);			// Set CS low to enable device
	SPI.transfer(regAddr, SPI_CONTINUE); // Write address over SPI bus
	SPI.transfer(0x00);				// Write 0x00 to the SPI bus fill the 16 bit transaction requirement
	digitalWrite(_CS, HIGH);		// Set CS high to disable device

	delayMicroseconds(WORD_DELAY);	// Delay to not violate read rate (40us) (min. 24 us for 1 MHz clock)

	// Read data from requested register
	digitalWrite(_CS, LOW);			// Set CS low to enable device
	int16_t _dataOut = SPI.transfer16(0x0000); // Send (0x00) and place data into variable
	digitalWrite(_CS, HIGH);		// Set CS high to disable device

	delayMicroseconds(WORD_DELAY);	// Delay to not violate read rate (40us)
	SPI.endTransaction();
	xSemaphoreGive(_SPImut);

	return(_dataOut);
}

////////////////////////////////////////////////////////////////////////////
// Writes one byte of data to the specified register over SPI.
// Returns 1 when complete.
////////////////////////////////////////////////////////////////////////////
// regAddr - address of register to be written
// regData - data to be written to the register
////////////////////////////////////////////////////////////////////////////
int ADIS16448::regWrite(uint8_t regAddr, int16_t regData) {
	// Write register address and data
	uint16_t addr = (((regAddr & 0x7F) | 0x80) << 8);				// Toggle sign bit, and check that the address is 8 bits. "Sign bit" is !R/W => 1 to write
	uint16_t lowWord = (addr | (regData & 0xFF));					// OR Register address (A) with data(D) (AADD)
	uint16_t highWord = ((addr | 0x100) | ((regData >> 8) & 0xFF)); // OR Register address with data and increment address

	xSemaphoreTake(_SPImut, portMAX_DELAY);
	SPI.beginTransaction(_ADIS_SPIsettings);

	// Write highWord to SPI bus
	digitalWrite(_CS, LOW);			// Set CS low to enable device	
	SPI.transfer16(highWord);		// Write high word to SPI bus
	digitalWrite(_CS, HIGH);		// Set CS high to disable device
	
	delayMicroseconds(WORD_DELAY);	// Delay to not violate read rate (40us)
	
	// Write lowWord to SPI bus
	digitalWrite(_CS, LOW);			// Set CS low to enable device
	SPI.transfer16(lowWord);		// Write low word to SPI bus
	digitalWrite(_CS, HIGH);		// Set CS high to disable device

	delayMicroseconds(WORD_DELAY);	// Delay to not violate read rate (40us)

	SPI.endTransaction();
	xSemaphoreGive(_SPImut);
	return(1);
}

float ADIS16448::readAccel(char axis, int8_t sign) {
	uint8_t sensorReg;
	if (axis == 'x' || axis == 'X')
		sensorReg = XACCL_OUT;
	else if (axis == 'y' || axis == 'Y')
		sensorReg = YACCL_OUT;
	else if (axis == 'z' || axis == 'Z')
		sensorReg = ZACCL_OUT;
	else
		return NAN;
	int16_t accVal = regRead(sensorReg);
	return(float)((accVal / 1200.0f) * GRAVITY_G * (float)sign);
}

float ADIS16448::readGyro(char axis) {
	uint8_t sensorReg;
	if (axis == 'x' || axis == 'X')
		sensorReg = XGYRO_OUT;
	else if (axis == 'y' || axis == 'Y')
		sensorReg = YGYRO_OUT;
	else if (axis == 'z' || axis == 'Z')
		sensorReg = ZGYRO_OUT;
	else
		return NAN;
	int16_t gyroVal = regRead(sensorReg);
	if (_gyroRange == GyroRange::PM1000 || _gyroRange == GyroRange::PM500 || _gyroRange == GyroRange::PM250)
		return(float)(gyroVal * 0.01 * (float)_gyroRange);
	else
		return NAN;
}

float ADIS16448::readMagnetometer(char axis) {
	uint8_t sensorReg;
	if (axis == 'x' || axis == 'X')
		sensorReg = XMAGN_OUT;
	else if (axis == 'y' || axis == 'Y')
		sensorReg = YMAGN_OUT;
	else if (axis == 'z' || axis == 'Z')
		sensorReg = ZMAGN_OUT;
	else
		return NAN;
	int16_t magVal = regRead(sensorReg);
	return (float)(magVal / 7.0);
}

float ADIS16448::readPressure() {
	uint16_t pressVal = regRead(BARO_OUT);	// Binary data format (not two's complement), only positive values.
	return (float)(pressVal / 50.0);
}

float ADIS16448::readTemp() {
	int16_t tempVal = regRead(TEMP_OUT);
	return (float)((tempVal * 0.07386) + 31.0);
}