// ADIS16448.h

#ifndef _ADIS16448_h
#define _ADIS16448_h

#if defined(ARDUINO) && ARDUINO >= 100
#include "arduino.h"
#else
#include "WProgram.h"
#endif

#include "FreeRTOSConfig.h" // Local config for the project, must be included before <FreeRTOS_ARM.h>.
#include <FreeRTOS_ARM.h>
#include <SPI.h>

// User Register Memory Map from Table 6
#define FLASH_CNT 0x00  //Flash memory write count
#define XGYRO_OUT 0x04  //X-axis gyroscope output
#define YGYRO_OUT 0x06  //Y-axis gyroscope output
#define ZGYRO_OUT 0x08  //Z-axis gyroscope output
#define XACCL_OUT 0x0A  //X-axis accelerometer output
#define YACCL_OUT 0x0C  //Y-axis accelerometer output
#define ZACCL_OUT 0x0E  //Z-axis accelerometer output
#define XMAGN_OUT 0X10  //X-axis magnetometer output
#define YMAGN_OUT 0x12  //Y-axis magnetometer output
#define ZMAGN_OUT 0x14  //Z-axis magnetometer output
#define BARO_OUT 0x16   //Barometer pressure measurement, high word
#define TEMP_OUT 0x18   //Temperature output
#define XGYRO_OFF 0x1A  //X-axis gyroscope bias offset factor
#define YGYRO_OFF 0x1C  //Y-axis gyroscope bias offset factor
#define ZGYRO_OFF 0x1E  //Z-axis gyroscope bias offset factor
#define XACCL_OFF 0x20  //X-axis acceleration bias offset factor
#define YACCL_OFF 0x22  //Y-axis acceleration bias offset factor
#define ZACCL_OFF 0x24  //Z-axis acceleration bias offset factor
#define XMAGN_HIC 0x26  //X-axis magnetometer, hard iron factor
#define YMAGN_HIC 0x28  //Y-axis magnetometer, hard iron factor
#define ZMAGN_HIC 0x2A  //Z-axis magnetometer, hard iron factor
#define XMAGN_SIC 0x2C  //X-axis magnetometer, soft iron factor
#define YMAGN_SIC 0x2E  //Y-axis magnetometer, soft iron factor
#define ZMAGN_SIC 0x30  //Z-axis magnetometer, soft iron factor
#define GPIO_CTRL 0x32  //GPIO control
#define MSC_CTRL 0x34   //MISC control
#define SMPL_PRD 0x36   //Sample clock/Decimation filter control
#define SENS_AVG 0x38   //Digital filter control
#define SEQ_CNT 0x3A    //MAGN_OUT and BARO_OUT counter
#define DIAG_STAT 0x3C  //System status
#define GLOB_CMD 0x3E   //System command
#define ALM_MAG1 0x40   //Alarm 1 amplitude threshold
#define ALM_MAG2 0x42   //Alarm 2 amplitude threshold
#define ALM_SMPL1 0x44  //Alarm 1 sample size
#define ALM_SMPL2 0x46  //Alarm 2 sample size
#define ALM_CTRL 0x48   //Alarm control
#define LOT_ID1 0x52    //Lot identification number
#define LOT_ID2 0x54    //Lot identification number
#define PROD_ID 0x56    //Product identifier
#define SERIAL_NUM 0x58 //Lot-specific serial number

// Gravitational constant
#define GRAVITY_G 9.81f

// SPI frequency
#define ADIS_SPI_FREQ 1750000UL

// Limit read rate to minimum 40us (Table 2 in datasheet), must be adjusted depending on SPI frequency
#define WORD_DELAY 32

// Gyro range in degrees per second
enum class GyroRange : uint8_t {
	PM250 = 0x01,
	PM500 = 0x02,
	PM1000 = 0x04
};


// ADIS16448 Class Definition
class ADIS16448 {

public:
	//ADIS16448 Constructor (ChipSelect, DataReady, Reset Pins)
	ADIS16448(int CS, int DR, int RST);

	// Initialize device
	void init(SemaphoreHandle_t SPImut);

	// Estimate gravity
	float calcGravity(char axis, int8_t sign);

	 //Read register (two bytes) Returns signed 16 bit data.
	int16_t regRead(uint8_t regAddr);

	//Write register (two bytes). Returns 1 when complete.
	int regWrite(uint8_t regAddr, int16_t regData);

	// Read acceleration data for the specified axis. Returns acceleration in m/s^2.
	float readAccel(char axis, int8_t sign);

	// Read gyroscope data for the specified axis. Returns roation in deg/sec.
	float readGyro(char axis);

	// Read magnetometer data for the specified axis. Returns magnetic field in mgauss.
	float readMagnetometer(char axis);

	// Read barometric pressure. Returns pressure in mbar.
	float readPressure();

	// Read internal temperature. Returns temperature in degC.
	float readTemp();

private:
	//Variables to store hardware pin assignments.
	int _CS;
	int _DR;
	int _RST;

	// Mutex for access to the SPI
	SemaphoreHandle_t _SPImut;

	// Settings for SPI communication
	SPISettings _ADIS_SPIsettings;

	GyroRange _gyroRange;
};

#endif

