// 
// 
// 

#include "EPOS.h"

bool EPOSClass::Init(uint32_t baudrate)
{
	if(!_CO.Init(1, baudrate)){
		return false;
	}
	_state = EPOSstate::Start;
	_status.remote = false;
	_CO.AttachListener(this);

	_CO.ResetNode();
	int retries = 0;
	while (_state != EPOSstate::SwitchOnDisabled) {
		_CO.SDOupload(getStatusword);
		delay(500);
		retries++;
		if (retries >= 3)
			return false;
	}
	
	_CO.StartRemoteNode();
	retries = 0;
	while (_state != EPOSstate::SwitchOnDisabled || !_status.remote) {
		delay(10);
		_CO.SendSynch();
		retries++;
		if (retries >= 10)
			return false;
	}

	Enable();
	
	_CO.SDOdownload(halt);
	_readyForHoming = false;
	SetMaxProfileVelocity(10000);
	SetMaxAcceleration(2000);

	setMode(OperationMode::Position);
	return true;
}

EPOSstate EPOSClass::GetState()
{
	return _state;
}

void EPOSClass::handleReceivedMessage(CO_PDO *msg)
{
	switch (msg->id) {
	case COB_ID::TxPDO3:
		// Get statusword from byte 0-1 and mask out the insignificant bits
		_state = (EPOSstate)(msg->data.asInt & 0x000000000000417f); 
		_status.warning = bitRead(msg->data.asBytes[0], 7);
		_status.remote = bitRead(msg->data.asBytes[1], 1);
		_status.targetReached = bitRead(msg->data.asBytes[1], 2);
		_status.homingAttained = bitRead(msg->data.asBytes[1], 4);
		_status.homingError = bitRead(msg->data.asBytes[1], 5);
		
		//Get position from byte 2-5
		_lastPosition = (msg->data.asInt & 0x0000ffffffff0000) >> 16;
		_positionTimeStamp = millis();
		break;
	case COB_ID::TxPDO4:
		_mode = (OperationMode)msg->data.asBytes[0];
		break;
	}
}

void EPOSClass::handleReceivedMessage(CO_SDO *msg)
{
	if (msg->index == 0x6041 && msg->subIndex == 0x00) {
		// Get statusword from byte 0-1 and mask out the insignificant bits
		_state = (EPOSstate)(msg->data.asInt & 0x0000417f);
		_status.warning = bitRead(msg->data.asBytes[0], 7);
		_status.remote = bitRead(msg->data.asBytes[1], 1);
		_status.targetReached = bitRead(msg->data.asBytes[1], 2);
		_status.homingAttained = bitRead(msg->data.asBytes[1], 4);
		_status.homingError = bitRead(msg->data.asBytes[1], 5);
	}
	else if (msg->index == 0x6040 && msg->subIndex == 0x00) {
		_readyForHoming = !bitRead(msg->data.asBytes[0], 4);
	}
}

void EPOSClass::setMode(OperationMode newMode)
{
	if (_mode != newMode) {
		_modesOfOperation.data.asBytes[0] = (uint8_t)newMode;
		_CO.SDOdownload(_modesOfOperation);
		while (_mode != newMode) {
			_CO.SendSynch();
			delay(10);
		}
	}
}

void EPOSClass::SetPosition(int32_t pos)
{
	_positionModeSettingValue.data.asInt = pos;
	_CO.SDOdownload(_positionModeSettingValue);
	_CO.SendSynch();
}

int32_t EPOSClass::GetPosition()
{
	while (millis() - _positionTimeStamp >= 2) {
		_CO.SendSynch();
		delayMicroseconds(100);
	}
	return _lastPosition;
}

void EPOSClass::SetMaxProfileVelocity(uint32_t maxVel)
{
	_maxProfileVelicity.data.asUInt = maxVel;
	_CO.SDOdownload(_maxProfileVelicity);
}

void EPOSClass::SetMaxAcceleration(uint32_t maxAcc)
{	
	_maxAcceleration.data.asUInt = maxAcc;
	_CO.SDOdownload(_maxAcceleration);
}

void EPOSClass::RunHoming()
{
	setMode(OperationMode::Homing);
	while (!_readyForHoming) {
		_CO.SDOdownload(enableOperation);
		delay(10);
	}
	_CO.SDOdownload(startHoming);
	while (!_status.homingAttained) {
		_CO.SendSynch();
		delay(10);
	}
	_readyForHoming = false;
	setMode(OperationMode::Position);
}

void EPOSClass::Enable()
{
	if (_state != EPOSstate::ReadyToSwitchOn) {
		_CO.SDOdownload(shutdown);
		while (_state != EPOSstate::ReadyToSwitchOn) {
			delay(10);
			_CO.SendSynch();
		}
	}
	_CO.SDOdownload(enableOperation);
	while (_state != EPOSstate::OperationEnable) {
		delay(10);
		_CO.SendSynch();
	}
}

void EPOSClass::Disable()
{
	_CO.SDOdownload(disableOperation);
	while (_state != EPOSstate::SwitchedOn) {
		delay(10);
		_CO.SendSynch();
	}
}

void EPOSClass::ResetFaults()
{
	_CO.SDOdownload(shutdown);
	_CO.SDOdownload(faultReset);
	while (_state != EPOSstate::SwitchOnDisabled) {
		delay(10);
		_CO.SendSynch();
	}
}