// EPOS.h

#ifndef _EPOS_h
#define _EPOS_h

#if defined(ARDUINO) && ARDUINO >= 100
	#include "arduino.h"
#else
	#include "WProgram.h"
#endif

#include "CANopen.h"

enum class OperationMode : int8_t {
	InterpolatedPosition = 7,
	Homing = 6,
	ProfileVelocity = 3,
	ProfilePosition = 1,
	Position = -1,
	Velocity = -2,
	Current = -3,
	Diagnostics = -4,
	MasterEncoder = -5,
	StepDirection = -6
};

enum class EPOSstate : uint16_t {
	Start = 0x0000,
	NotReadyToSwitchOn = 0x0100,
	SwitchOnDisabled = 0x0140,
	ReadyToSwitchOn = 0x0121,
	SwitchedOn = 0x0123,
	Refresh = 0x4123,
	MeasureInit = 0x4133,
	OperationEnable = 0x0137,
	QuickstopEnable = 0x0117,
	FaultReactionActiveDisabled = 0x010f,
	FaultReactionActiveEnabled = 0x011f,
	Fault = 0x0108
};

//---------------------------------------default messages----------------------------------------------
//********************** uint8_t length | uint16_t index | uint8_t subindex| uint8_t data[4](LSB...MSB)
static const CO_SDO disableOperation = { 2, 0x6040, 0x00,{ 0x07, 0, 0, 0 } };
static const CO_SDO shutdown = { 2, 0x6040, 0x00,{ 0x06, 0, 0, 0 } };
static const CO_SDO enableOperation = { 2, 0x6040, 0x00,{ 0x0f, 0, 0, 0 } };
static const CO_SDO halt = { 2, 0x6040, 0x00,{ 0x0f, 0x01, 0, 0 } };
static const CO_SDO faultReset = { 2, 0x6040, 0x00,{ 0x80, 0, 0, 0 } };
static const CO_SDO startHoming = { 2, 0x6040, 0x00,{ 0x1F, 0, 0, 0 } };
static const CO_SDO getStatusword = { 0, 0x6041, 0x00,{ 0, 0, 0, 0 } };
static const CO_SDO homeOffset = { 4, 0x607C, 0x00,{ 0, 0, 0, 0 } };
static const CO_SDO minSoftwarePositionLimit = { 4, 0x607D, 0x01,{ 0, 0, 0, 0 } };
static const CO_SDO maxSoftwarePositionLimit = { 4, 0x607D, 0x02,{ 0, 0, 0, 0 } };

class EPOSClass : CANopenListener
{
private:
		volatile struct EPOSstatus {
			bool warning;
			bool remote;
			bool targetReached;
			bool homingAttained;
			bool homingError;
		} _status;

		CANopen _CO;
		volatile EPOSstate _state;
		volatile OperationMode _mode;
		volatile bool _targetReached;
		volatile bool _homingAttained;
		volatile bool _homingError;
		volatile bool _readyForHoming;
		volatile int32_t _lastPosition;
		volatile uint32_t _positionTimeStamp;
		CO_SDO _positionModeSettingValue = { 4, 0x2062, 0x00,(Data32){ 0, 0, 0, 0 } };
		CO_SDO _modesOfOperation = { 1, 0x6060, 0x00,(Data32) { 0, 0, 0, 0 } };
		CO_SDO _maxProfileVelicity = { 4, 0x607F, 0x00,(Data32) { 0, 0, 0, 0 } };
		CO_SDO _maxAcceleration = { 4, 0x60C5, 0x00,(Data32) { 0, 0, 0, 0 } };

		void handleReceivedMessage(CO_PDO *msg);
		void handleReceivedMessage(CO_SDO *msg);
		void setMode(OperationMode newMode);
		
public:
		bool Init(uint32_t baudrate);

		EPOSstate GetState();
		
		void SetPosition(int32_t pos);
		int32_t GetPosition();

		void SetMaxProfileVelocity(uint32_t maxVel);
		void SetMaxAcceleration(uint32_t maxAcc);

		void RunHoming();
			
		void Enable();
		void Disable();
		
		void ResetFaults();
};

#endif

