// 
// 
// 

#include "HLController.h"

bool HLControllerClass::Init(QueueHandle_t filterParamQueue, QueueHandle_t t2maxSettlingTimeQueue, SemaphoreHandle_t SPImut)
{
	// Setup CAN, CANopen and EPOS2
	if (!_motorController.Init(1000000)) {
		HLSerial.Send(OutGoingCmd::errorGeneral, "Failed to initialize motor controller");
		return false;
	}
	_adc.init(SPImut, 16);
	calculateParameters();
	sendStatus();
	_filterParamQueue = filterParamQueue;
	_t2maxSettlingTimeQueue = t2maxSettlingTimeQueue;
	_currentData.valveSetpoint = 95.0f;
	return true;
}

void HLControllerClass::Run(float v_hat)
{
	if (HLSerial.Read(_incomingMsg))
		HLController.checkProtocol(_incomingMsg);

	if (!_manualMode) {
		readProcess(_process);

		_currentData.estimatedVelocity = v_hat;

		// Calculate desired flow rate
		float desiredFlow = simpleFeedForwardController(v_hat);

		// Calculate desired valve position
		float desiredPosition = inverseChokeChar(desiredFlow, _process.pressureUpstreamHL, _process.pressureDownstreamHL);
		
		if (desiredPosition > 0.0f) {
			_currentData.valveSetpoint = desiredPosition;
			if (desiredPosition > 100.0f)
				_currentData.valveSetpoint = 100.0f;
		}
	}
	// Send desired valve position to motor controller
	if (_motorController.GetState() == EPOSstate::OperationEnable && _homingComplete) {
		if (_currentData.valveSetpoint > (_prevSetpoint + _MIN_CHANGE) || _currentData.valveSetpoint < (_prevSetpoint - _MIN_CHANGE)) {
			_motorController.SetPosition(scaleSP(_currentData.valveSetpoint));
			_prevSetpoint = _currentData.valveSetpoint;
		}
	}
	_currentData.valvePosition = scaleTravel(_motorController.GetPosition());
	HLSerial.Send(_currentData);
}

void HLControllerClass::calculateParameters() {
	/*
	// Cross - sectional area of drill string(m^2)
	_param.Ad = PI*sq((_param.dd / 2));
	// Cross - sectional area of annulus(m^2)
	_param.Aa = PI*(sq(_param.dhole) - sq(_param.dBHA)) / 4;
	// Volume of the hole(m^3)
	_param.Vhole = PI*sq(_param.dhole) / 4 * _param.Lhole - PI*sq(_param.dBHA) / 4 * _param.LBHA - PI*(sq(_param.durod) + sq(_param.dlrod)) / 4 * (_param.Lhole - _param.LBHA) / 2;
	*/
	// Area of the lower end of BHA(m^2)
	_param.Al = PI*(sq(_param.dBHA) - sq(_param.dlrod)) / 4;
	// Area of the upper end of BhA(m^2)
	_param.Au = PI*(sq(_param.dBHA) - sq(_param.durod)) / 4;
	/*
	// Effective bulk modulus in drill string(Pa)
	_param.betad = 1 / (1 / _param.betaf + 1 / _param.Ed*_param.dd / _param.taud);
	// Effective bulk modulus in annulus(Pa)
	_param.betaa = 1 / (1 / _param.betaf + 1 / _param.Ea*_param.da / _param.taua);
	// Effective bulk modulus downhole(Pa)
	_param.betadh = _param.betaf;
	*/
}

void HLControllerClass::checkProtocol(HeaveLockMessage input) {
	switch (input.cmd)
	{
	case IncomingCmd::closeValve:
		if (_homingComplete) {
			_currentData.valveSetpoint = 0.0f;
			HLSerial.Send(OutGoingCmd::valveClosed);
		}
		else
			HLSerial.Send(OutGoingCmd::errorGeneral, "Homing needs to be executed");
		break;
	case IncomingCmd::openValve:
		if (_homingComplete) {
			_currentData.valveSetpoint = 100.0f;
			HLSerial.Send(OutGoingCmd::valveOpened);
		}
		else
			HLSerial.Send(OutGoingCmd::errorGeneral, "Homing needs to be executed");
		break;
	case IncomingCmd::startHeaveComp:
		if (_homingComplete) {
			_manualMode = false;
			HLSerial.Send(OutGoingCmd::startHeaveCompConfirmed);
		}
		else {
			HLSerial.Send(OutGoingCmd::errorGeneral, "Homing needs to be executed");
		}
		break;
	case IncomingCmd::stopHeaveComp:
		_manualMode = true;
		HLSerial.Send(OutGoingCmd::stopHeaveCompConfirmed);
		break;
	case IncomingCmd::errorAck:
		HLSerial.Send(OutGoingCmd::errorAckConfirmed);
		break;
	case IncomingCmd::startHoming:
		HLSerial.Send(OutGoingCmd::homingStarted);
		_homingComplete = false;
		_motorController.RunHoming();
		_homingComplete = true;
		HLSerial.Send(OutGoingCmd::homingComplete);
		break;
	case IncomingCmd::enableValveController:
		_motorController.Enable();
		HLSerial.Send(OutGoingCmd::valveControllerEnabled);
		break;
	case IncomingCmd::disableValveController:
		_motorController.Disable();
		HLSerial.Send(OutGoingCmd::valveControllerDisabled);
		break;
	case IncomingCmd::setValveSetPoint:
		float tempSP;
		tempSP = HLSerial.ConvertBytesToFloat(input);
		if (tempSP >= 0.0f && tempSP <= 100.0f) {
			_currentData.valveSetpoint = tempSP;
			HLSerial.Send(OutGoingCmd::valveSetpointChanged);
		}
		else
		{
			HLSerial.Send(OutGoingCmd::errorGeneral, "Invalid setpoint");
		}
		break;
	case IncomingCmd::getStatus:
		sendStatus();
		break;
	case IncomingCmd::setAllFiltPhaseParams:
		FilterParam newParams;
		getFilterParamsFromHLMessage(input, newParams);
		xQueueOverwrite(_filterParamQueue, &newParams);
		HLSerial.Send(newParams);
		break;
	case IncomingCmd::get_valPos_valSet_estVel:
		_currentData.valvePosition = scaleTravel(_motorController.GetPosition());
		HLSerial.Send(_currentData);
		break;
	case IncomingCmd::setSPfromSimulink:
		float tempSP2;
		tempSP2 = HLSerial.ConvertBytesToFloat(input);
		if (tempSP2 >= 0.0f && tempSP2 <= 100.0f) {
			_currentData.valveSetpoint = tempSP2;
		}
		else
		{
			HLSerial.Send(OutGoingCmd::errorGeneral, "Invalid setpoint");
		}
		break;
	case IncomingCmd::setT2max:
		float tempT2max;
		tempT2max = HLSerial.ConvertBytesToFloat(input);
		xQueueOverwrite(_t2maxSettlingTimeQueue, &tempT2max);
		HLSerial.Send(OutGoingCmd::t2maxSetConfirmed);
		break;
	case IncomingCmd::setPumpRate:		
		_param.initqp = HLSerial.ConvertBytesToFloat(input);		
		HLSerial.Send(OutGoingCmd::pumpRateSetConfirmed,_param.initqp);
		break;
	case IncomingCmd::setMaxMotorAcceleration:		
		uint32_t tempMaxAcc;
		tempMaxAcc = HLSerial.ConvertBytesToUint32(input);
		_motorController.SetMaxAcceleration(tempMaxAcc);
		HLSerial.Send(OutGoingCmd::MaxMotorAccelerationChanged);
		break;
	default:
		HLSerial.Send(OutGoingCmd::errorGeneral, "Invalid command received");
		break;
	}
}

void HLControllerClass::sendStatus() {	
	if (_motorController.GetState() == EPOSstate::OperationEnable)
		HLSerial.Send(OutGoingCmd::valveControllerEnabled);
	else if (_motorController.GetState() == EPOSstate::SwitchedOn)
		HLSerial.Send(OutGoingCmd::valveControllerDisabled);
	else
		HLSerial.Send(OutGoingCmd::errorGeneral, "Unknown motor controller state");	

	if (_manualMode)
		HLSerial.Send(OutGoingCmd::stopHeaveCompConfirmed);
	else
		HLSerial.Send(OutGoingCmd::startHeaveCompConfirmed);

	if (_homingComplete)
		HLSerial.Send(OutGoingCmd::homingComplete);
	else
		HLSerial.Send(OutGoingCmd::homingNotExecuted);

	HLSerial.Send(OutGoingCmd::pumpRateSetConfirmed, _param.initqp);
}

void HLControllerClass::readProcess(ProcessMeasurements & measurements)
{
	// Read values from ADC
	_adc.analogReadConfigNext(0, SignalType::DIFFERENTIAL, Polarity::UNIPOLAR, Range::RANGE5V);
	float voltageCh0 = _adc.analogReadConfigNext(2, SignalType::DIFFERENTIAL, Polarity::UNIPOLAR, Range::RANGE5V);
	float voltageCh2 = _adc.analogReadConfigNext(4, SignalType::DIFFERENTIAL, Polarity::UNIPOLAR, Range::RANGE5V);
	float voltageCh4 = _adc.analogReadConfigNext(6, SignalType::DIFFERENTIAL, Polarity::UNIPOLAR, Range::RANGE5V);
	float voltageCh6 = _adc.analogReadConfigNext(0, SignalType::DIFFERENTIAL, Polarity::UNIPOLAR, Range::RANGE5V);

	measurements.flowUpstreamHL = scaleToEngineeringValues(voltageCh0, 0.0f, 30.0f); // [l/min]
	measurements.pressureUpstreamHL = scaleToEngineeringValues(voltageCh2, 0.0f, 70.0f); // [bar]
	measurements.pressureDownhole = scaleToEngineeringValues(voltageCh4, 0.0f, 25.0f); // [bar]
	measurements.pressureDownstreamHL = scaleToEngineeringValues(voltageCh6, 0.0f, 25.0f); // [bar]
}

float HLControllerClass::scaleToEngineeringValues(float val, float eMin, float eMax) {
	static float vMin = 1.0f;
	static float vMax = 5.0f;

	return ((val - vMin) / (vMax - vMin)) * (eMax - eMin) + eMin;
}

float HLControllerClass::simpleFeedForwardController(float v_hat)
{
	float desiredFlow = (_param.Al - _param.Au)*60000.0f*v_hat + _param.initqp;	// 60000.0f converts from m^3/s to l/min
	return desiredFlow;
}

float HLControllerClass::inverseChokeChar(float desiredFlow, float pressureUpstream, float pressureDownstream)
{
	// Choke characteristic parameters
	static float a = 0.0715f;
	static float b = 0.5958f;
	static float c = 0.2954f;
	static float d = 0.737f;
	static float e = 0.6236f; // Bias because the valve is never fully sealed
	float desiredValvePos = -1.0f;

	float deltaP = pressureUpstream - pressureDownstream;

	// Based on the choke characteristic: a*q^2 + b*q = (c*u^2 + d*u + e)*p_delta
	// Results in an inverse choke characteristic: u = (-d / (2 * c)) + sqrt(sq(d) - 4 * c*(e - (a*sq(desiredFlow) + b*desiredFlow) / deltaP)) / 2*c

	static float bulkConstant1 = (-d / (2 * c)); // Done once to enhance calc. speed
	static float bulkConstant2 = sq(d) / (4 * sq(c)) - e / c; // Done once to enhance calc. speed

	if (deltaP > 0.0f) {
		desiredValvePos = bulkConstant1 + sqrt(bulkConstant2 + (a*sq(desiredFlow) + b*desiredFlow) / (deltaP*c));
	}
	
	return desiredValvePos*100;	// desiredValvePos is a value [0,1]. This program operates with setpoint [0,100]
}

void HLControllerClass::getFilterParamsFromHLMessage(HeaveLockMessage input, FilterParam &newParams) {
	union Value
	{
		struct {
			byte asBytes[16];
			bool resetBuffer;
		};
		FilterParam par;
	} value;
	value.resetBuffer = true;
	for (int i = 0; i < 16; i++)
	{
		value.asBytes[i] = input.bData[i];
	}
	newParams = value.par;
}

int32_t HLControllerClass::scaleSP(float input) {
	return (int32_t)((input / 100.0f) * (float)_MAX_POS);
}

float HLControllerClass::scaleTravel(int32_t input) {
	return ((float)input / _MAX_POS) * 100.0f;
}

HLControllerClass HLController;
