// HLController.h

#ifndef _HLCONTROLLER_h
#define _HLCONTROLLER_h

#if defined(ARDUINO) && ARDUINO >= 100
	#include "arduino.h"
#else
	#include "WProgram.h"
#endif

#include "FreeRTOSConfig.h" // Local config for the project, must be included before <FreeRTOS_ARM.h>.
#include <FreeRTOS_ARM.h>

#include "EPOS.h"
#include "ExtendedADCShield.h"
#include "FilterParam.h"
#include "HLSerial.h"
#include "TelemetryData.h"

struct LabParameters {
		// TODO: Verify parameters in new lab
		/*
		// Common parameters.
		float rho = 998.2;            // Density(kg / m3) : water is default
		float mu = 0.001;             // Viscosity(sPa) : water is default
		float betaf = 2.2e9;          // Bulk modulus(Pa) : water is default
		float g = 9.81;               // Aceleration of gravity(m / s^2)
		float p0 = 1;                 // Pressure downstream choke(bar) : Atmospheric is default
		float Lhole = 0.8;            // Length of down - hole section(m)
		float epsr = 0;               // Roughness for downhole material(m)

		// Parameters for annulus
		float La = 900;               // Length of pipe(m)
		float da = 0.016;             // Pipe inner diameter(m)
		float Da = 2.13;              // Diameter of pipe coil(m)
		float thetaa = 1.5733;        // Inclination of pipe(radians)
		float Ea = 1.17e11;           // Youngs modulus of pipe(Pa) : copper is default
		float taua = 0.0015;          // Wall thickness of pipe(m)
		float epsa = 0;               // Roughness of pipe(m) ?

		// Parameters for drillstring
		float Ld = 900;               // Length of pipe(m)
		float dd = 0.016;             // Pipe inner diameter(m)
		float Dd = 2.13;              // Diameter of pipe coil(m)
		float thetad = 1.5733;        // Inclination of pipe(radians)
		float Ed = 1.17e11;           // Youngs modulus of pipe(Pa) : copper is default
		float taud = 0.0015;          // Wall thickness of pipe(m)
		float epsd = 0;               // Roughness of pipe(m) ?
		*/

		// Design variables
		float initqp = 15;            // Pump volumetric flow at simulation start(l / min)
		//float dhole = 0.060;          // Diameter of hole(m)
		float dBHA = 0.054;           // Diameter of BHA(m)
		//float LBHA = 0.40;            // Length of BHA(m)
		float dlrod = 0.020;          // Diameter of lower rod(m)
		float durod = 0.032;          // Diameter of upper rod(m)
		//float Kbit = 3.7e-6;          // Discharge coefficient for the bit(m3 / s / sqrt(Pa))

		/*
		// Cross - sectional area of drill string(m^2)
		float Ad;
		// Cross - sectional area of annulus(m^2)
		float Aa;
		// Volume of the hole(m^3)
		float Vhole;
		*/
		// Area of the lower end of BHA(m^2)
		float Al;
		// Area of the upper end of BhA(m^2)
		float Au;
		/*
		// Effective bulk modulus in drill string(Pa)
		float betad;
		// Effective bulk modulus in annulus(Pa)
		float betaa;
		// Effective bulk modulus downhole(Pa)
		float betadh;
		*/
};

struct ProcessMeasurements {
	float flowUpstreamHL;		// FT2
	float pressureUpstreamHL;	// PT6
	float pressureDownhole;		// PT8
	float pressureDownstreamHL;	// PT9
};

class HLControllerClass
{
 private:
	 static const int32_t _MAX_POS = 16000;
	 const float _MIN_CHANGE = 0.05f;
	 float _prevSetpoint = 0.0f;
	 bool _manualMode = true;
	 bool _homingComplete = false;
	 TelemetryData _currentData;
	 EPOSClass _motorController;
	 HeaveLockMessage _incomingMsg;
	 ProcessMeasurements _process;
	 ExtendedADCShield _adc;
	 LabParameters _param;
	 QueueHandle_t _filterParamQueue;
	 QueueHandle_t _t2maxSettlingTimeQueue;

	 void calculateParameters();
	 void checkProtocol(HeaveLockMessage input);
	 void sendStatus();
	 void readProcess(ProcessMeasurements &measurements);
	 float scaleToEngineeringValues(float val, float eMin, float eMax);
	 float simpleFeedForwardController(float v_hat);
	 float inverseChokeChar(float desiredFlow, float pressureUpstream, float pressureDownstream);

	 static void getFilterParamsFromHLMessage(HeaveLockMessage input, FilterParam & newParams);
	 static int32_t scaleSP(float input);
	 static float scaleTravel(int32_t input);

 public:
	 bool Init(QueueHandle_t filterParamQueue, QueueHandle_t t2maxSettlingTimeQueue, SemaphoreHandle_t SPImut);
	 void Run(float v_hat);
};

extern HLControllerClass HLController;

#endif

