
#include "FreeRTOSConfig.h" // Local config for the project, must be included before <FreeRTOS_ARM.h>.
#include <FreeRTOS_ARM.h>

#include "ADIS16448.h"
#include "Buffer.h"
#include "FilterParam.h"
#include "HeaveFilter.h"
#include "HLController.h"
#include "HLSerial.h"
#include "ParamEstimation.h"
#include "TaskParam.h"

#define ASSERT_LED_PIN 13
#define RXLED	72
#define TXLED	73

#define IMU_DATA_READY 2
#define IMU_RESET 3
#define IMU_CS 4

#define ACC_AXIS 'y'
#define ACC_SIGN -1 // 1 for positive and -1 for negative

// Task parameters
ParEstTaskParam paramEstArgs;
MainProgramTaskParam mainArgs;
ControllerTaskParam contrArgs;
ReadAccTaskParam readAccArgs;

TaskHandle_t velocityEstimationTaskHandle;
TaskHandle_t readAccTaskHandle;
//------------------------------------------------------------------------------
void setup() {
	pinMode(ASSERT_LED_PIN, OUTPUT);
	pinMode(RXLED, OUTPUT);
	pinMode(TXLED, OUTPUT);
	digitalWrite(ASSERT_LED_PIN, LOW);
	blinky();
	HLSerial.Begin(238636);
	HLSerial.Send(OutGoingCmd::infoMessage, "Initializing...");	

	// Create task at lowest priority
	paramEstArgs.parameterQueue = xQueueCreate(1, sizeof(FilterParam));
	paramEstArgs.bufferQueue = xQueueCreate(1, sizeof(Buffer));
	paramEstArgs.t2maxSettlingTimeQueue = xQueueCreate(1, sizeof(float));
	BaseType_t s1 = xTaskCreate(ParamEstimation::calculateFilterParametersTask, "ParEst", configMINIMAL_STACK_SIZE + 11000, &paramEstArgs, tskIDLE_PRIORITY + 1, NULL);

	mainArgs.parameterQueue = paramEstArgs.parameterQueue;
	mainArgs.bufferQueue = paramEstArgs.bufferQueue;
	mainArgs.estVelocityQueue = xQueueCreate(1, sizeof(float));
	mainArgs.accDataQueue = xQueueCreate(1, sizeof(float));
	mainArgs.watchDogQueue = xQueueCreate(1, sizeof(FilterParam));
	BaseType_t s2 = xTaskCreate(velocityEstimationTask, "VelEst", configMINIMAL_STACK_SIZE + 2500, &mainArgs, tskIDLE_PRIORITY + 2, &velocityEstimationTaskHandle);

	contrArgs.parameterQueue = paramEstArgs.parameterQueue;
	contrArgs.estVelocityQueue = mainArgs.estVelocityQueue;
	contrArgs.t2maxSettlingTimeQueue = paramEstArgs.t2maxSettlingTimeQueue;
	contrArgs.SPImut = xSemaphoreCreateMutex();
	BaseType_t s3 = xTaskCreate(controllerTask, "Contr", configMINIMAL_STACK_SIZE + 200, &contrArgs, tskIDLE_PRIORITY + 3, NULL);

	readAccArgs.SPImut = contrArgs.SPImut;
	readAccArgs.accDataQueue = mainArgs.accDataQueue;
	BaseType_t s4 = xTaskCreate(readAccTask, "ReadAcc", configMINIMAL_STACK_SIZE + 100, &readAccArgs, tskIDLE_PRIORITY + 4, &readAccTaskHandle);
		
	// Check for creation errors
	if (paramEstArgs.parameterQueue == NULL || paramEstArgs.bufferQueue == NULL || mainArgs.estVelocityQueue == NULL || mainArgs.accDataQueue == NULL || mainArgs.watchDogQueue == NULL
		|| velocityEstimationTaskHandle == NULL || readAccTaskHandle == NULL || contrArgs.SPImut == NULL || s1 != pdPASS || s2 != pdPASS || s3 != pdPASS || s4 != pdPASS) { 
		HLSerial.Send(OutGoingCmd::errorGeneral, "Creation problem");
		while (1);
	}

	HLSerial.Send(OutGoingCmd::infoMessage, "Starting tasks...");
	vTaskStartScheduler();	// Should not return
	HLSerial.Send(OutGoingCmd::infoMessage, "Insufficient RAM");
	while (1);
}
//------------------------------------------------------------------------------
// WARNING idle loop has a very small stack (configMINIMAL_STACK_SIZE)
// loop must never block
void loop() {
	// Not used.
}

static void velocityEstimationTask(void *pvParameters)
{
	MainProgramTaskParam *mainArgs = (MainProgramTaskParam*)pvParameters;
	QueueHandle_t paramQueue = mainArgs->parameterQueue;
	QueueHandle_t bufferQueue = mainArgs->bufferQueue;
	QueueHandle_t estVelocityQueue = mainArgs->estVelocityQueue;
	QueueHandle_t accDataQueue = mainArgs->accDataQueue;
	QueueHandle_t watchDogQueue = mainArgs->watchDogQueue;
	float sampleTimeAcc = mainArgs->sampleTimeAcc;
	float sampleTimeFFT = mainArgs->sampleTimeFFT;
	HLSerial.Send(OutGoingCmd::infoMessage, "velocityEstimationTask started...");

	FilterParam filterParameters;
	Buffer accBuffer;
	HeaveFilter heaveFilter;
	float accReading = 0.0;
	float v_hat = 0.0;
	int sampleNo = 0;
	//unsigned long maxTime = 0;
	bool toggleLedOn = false;
	bool paramsCalculated = false;
	const int FFTSampleRatio = (int)(sampleTimeFFT / sampleTimeAcc); //evaluates to 10 as of march 2016

	ulTaskNotifyTake(pdTRUE, portMAX_DELAY);	// Wait until IMU is ready

	TickType_t xLastWakeTime = xTaskGetTickCount();
	// Attach interrupt, adjusting interrupt priority according to http://www.freertos.org/RTOS-Cortex-M3-M4.html
	attachInterrupt(digitalPinToInterrupt(IMU_DATA_READY), accDataReadyHandler, RISING);
	NVIC_SetPriorityForAllInterrupts(configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY);
	while (1) {
		if (xQueueReceive(accDataQueue, &accReading, portMAX_DELAY) != pdTRUE) {
			HLSerial.Send(OutGoingCmd::errorGeneral, "Did not receive accData successfully");
			continue;
		}
		sampleNo++;

		if (sampleNo >= FFTSampleRatio) {
			BufferStatus status = accBuffer.addData(accReading);
			sampleNo = 0;
			if (status == BufferStatus::FULL) {
				if (xQueueSendToFront(bufferQueue, &accBuffer, portMAX_DELAY) != pdTRUE) {
					HLSerial.Send(OutGoingCmd::errorGeneral, "Could not send accBuffer to queue.");
				}
				accBuffer.reset();
			}

			if (toggleLedOn) {
				digitalWrite(TXLED, HIGH);
				toggleLedOn = false;
			}
			else {
				digitalWrite(TXLED, LOW);
				toggleLedOn = true;
			}
		}

		if (uxQueueMessagesWaiting(paramQueue) > 0) {
			FilterParam newPar;
			if (xQueueReceive(paramQueue, &newPar, 0) == pdTRUE) {
				filterParameters = newPar;
			}
			if (newPar.resetBuffer)
				accBuffer.reset();
			HLSerial.Send(filterParameters);
			paramsCalculated = true;
		}

		if (paramsCalculated) {
			v_hat = heaveFilter.estimateVelocity(accReading, filterParameters.w_c, sampleTimeAcc);
			v_hat = heaveFilter.phaseCorrection(v_hat, filterParameters, sampleTimeAcc);
		}
		else {
			v_hat = 0.0;
		}
		// Queue new v_hat to HLController-task
		// HLController sends new position to motorcontroller
		xQueueOverwrite(estVelocityQueue, &v_hat);

		xQueueOverwrite(watchDogQueue, &filterParameters);

		vTaskDelayUntil(&xLastWakeTime, (sampleTimeAcc*configTICK_RATE_HZ));
	}
}

static void accDataReadyHandler() {
	static signed portBASE_TYPE xHigherPriorityTaskWoken;
	xHigherPriorityTaskWoken = pdFALSE;
	vTaskNotifyGiveFromISR(readAccTaskHandle, &xHigherPriorityTaskWoken);
	portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}

static void readAccTask(void *pvParameters) {
	ReadAccTaskParam *readAccArgs = (ReadAccTaskParam*)pvParameters;
	SemaphoreHandle_t SPImut = readAccArgs->SPImut;
	QueueHandle_t accDataQueue = readAccArgs->accDataQueue;
	HLSerial.Send(OutGoingCmd::infoMessage, "readAccTask started...");

	ADIS16448 IMU(IMU_CS, IMU_DATA_READY, IMU_RESET);
	IMU.init(SPImut);
	vTaskDelay(pdMS_TO_TICKS(1000));	// Need time to start up, depending on sample rate and filter settings	
	float gravity = IMU.calcGravity(ACC_AXIS, ACC_SIGN);
	float accReading = IMU.readAccel(ACC_AXIS, ACC_SIGN) - gravity;

	xTaskNotifyGive(velocityEstimationTaskHandle);	// Notify velocity estimation task that startup is finished
	ulTaskNotifyTake(pdTRUE, portMAX_DELAY);		// Wait for first interrupt

	while (1) {
		if (ulTaskNotifyTake(pdTRUE, pdMS_TO_TICKS(100)) == 0) {
			HLSerial.Send(OutGoingCmd::errorGeneral, "readAcc timed out...");
			continue;
		}
		accReading = IMU.readAccel(ACC_AXIS, ACC_SIGN) - gravity;
		xQueueOverwrite(accDataQueue, &accReading);
	}
}

static void controllerTask(void *pvParameters) {
	ControllerTaskParam *HLContrArgs = (ControllerTaskParam*)pvParameters;
	QueueHandle_t paramQueue = HLContrArgs->parameterQueue;
	QueueHandle_t estVelocityQueue = HLContrArgs->estVelocityQueue;
	QueueHandle_t t2maxSettlingTimeQueue = HLContrArgs->t2maxSettlingTimeQueue;
	SemaphoreHandle_t SPImut = HLContrArgs->SPImut;

	HLSerial.Send(OutGoingCmd::infoMessage, "HLControllerTask started...");
	float v_hat = 0;
	while (!HLController.Init(paramQueue, t2maxSettlingTimeQueue, SPImut)) { // If failed, try again. Most likely no power to EPOS.
		vTaskDelay(pdMS_TO_TICKS(1000));
	}

	// Set priority for the CAN interrupt to a level that fits the FreeRTOS tasks in the system
	uint32_t can_priority = 6;
	NVIC_DisableIRQ(CAN0_IRQn);
	NVIC_ClearPendingIRQ(CAN0_IRQn);
	NVIC_SetPriority(CAN0_IRQn, ((uint32_t)15 - can_priority));
	NVIC_EnableIRQ(CAN0_IRQn);

	TickType_t xLastWakeTime = xTaskGetTickCount();
	while (true)
	{
		// Read estimated velocity from queue
		if (xQueueReceive(estVelocityQueue, &v_hat, portMAX_DELAY) != pdTRUE) {
			HLSerial.Send(OutGoingCmd::errorGeneral, "Did not receive v_hat successfully");
			continue;
		}
		HLController.Run(v_hat);
	}
}


void blinky() {
	digitalWrite(TXLED, LOW);
	delay(200);
	digitalWrite(TXLED, HIGH);
	delay(200);
	digitalWrite(TXLED, LOW);
	delay(200);
	digitalWrite(TXLED, HIGH);
	delay(200);
	digitalWrite(TXLED, LOW);
	delay(200);
	digitalWrite(TXLED, HIGH);
	delay(200);
}

void NVIC_SetPriorityForAllInterrupts(uint32_t priority) {
	NVIC_DisableIRQ(PIOA_IRQn);
	NVIC_ClearPendingIRQ(PIOA_IRQn);
	NVIC_SetPriority(PIOA_IRQn, priority);
	NVIC_EnableIRQ(PIOA_IRQn);

	NVIC_DisableIRQ(PIOB_IRQn);
	NVIC_ClearPendingIRQ(PIOB_IRQn);
	NVIC_SetPriority(PIOB_IRQn, priority);
	NVIC_EnableIRQ(PIOB_IRQn);

	NVIC_DisableIRQ(PIOC_IRQn);
	NVIC_ClearPendingIRQ(PIOC_IRQn);
	NVIC_SetPriority(PIOC_IRQn, priority);
	NVIC_EnableIRQ(PIOC_IRQn);

	NVIC_DisableIRQ(PIOD_IRQn);
	NVIC_ClearPendingIRQ(PIOD_IRQn);
	NVIC_SetPriority(PIOD_IRQn, priority);
	NVIC_EnableIRQ(PIOD_IRQn);
}