// 
// 
// 
#include "ParamEstimation.h"

#define ARM_MATH_CM3
#include <arm_math.h>
#include "Buffer.h"
#include "FilterParam.h"
#include "FreeRTOSConfig.h" // Local config for the project, must be included before <FreeRTOS_ARM.h>.
#include <FreeRTOS_ARM.h>
#include "Hamming.h"
#include "HLSerial.h"
#include "TaskParam.h"

#define MIN_WAVE_HEIGHT 0.1f	// [m]

void ParamEstimation::calculateFilterParametersTask(void *pvParameters)
{
	HLSerial.Send(OutGoingCmd::infoMessage, "calculateFilterParametersTask started...");
	ParEstTaskParam *args = (ParEstTaskParam*)pvParameters;
	QueueHandle_t paramQueue = args->parameterQueue;
	QueueHandle_t bufferQueue = args->bufferQueue;
	QueueHandle_t t2maxSettlingTimeQueue = args->t2maxSettlingTimeQueue;
	float sampleTimeFFT = args->sampleTimeFFT;
	float sampleTimeAcc = args->sampleTimeAcc;
	float dw = args->dw;
	float T_2max = args->T_2max;
	float sigma2 = args->sigma2;
	float S_nn = args->S_nn;

	Buffer seaStateEstBuffer;
	float windowedAccArray[BUFFER_LENGTH];
	float amplitudeArray[BUFFER_LENGTH];
	FilterParam params;
	params.resetBuffer = false;
	while (1)
	{
		if (xQueueReceive(bufferQueue, &seaStateEstBuffer, portMAX_DELAY) != pdTRUE) {
			HLSerial.Send(OutGoingCmd::errorGeneral, "Failed to receive new acceleration data buffer");
			continue;
		}
		if (uxQueueMessagesWaiting(t2maxSettlingTimeQueue) > 0) { //check if a new t_2max has been received
			float newT_2max;
			if (xQueueReceive(t2maxSettlingTimeQueue, &newT_2max, 0) == pdTRUE) {
				T_2max = newT_2max;
			}
		}
		// Sea state estimation:
		arm_mult_f32(seaStateEstBuffer.getData(), hammingWind, windowedAccArray, BUFFER_LENGTH);	// Applying Hamming window
		RealFFT(windowedAccArray, amplitudeArray);
		float w = getDominantWaveFreq(amplitudeArray, sampleTimeFFT);
		float A = getDominantWaveAmplitude(w, seaStateEstBuffer.getData(), sigma2);

		if (A >= MIN_WAVE_HEIGHT)
		{
			getFilterParams(w, A, dw, sampleTimeAcc, sigma2, T_2max, S_nn, &params);
			// Queue parameters
			xQueueSend(paramQueue, &params, portMAX_DELAY);
		}
	}
}

float ParamEstimation::getDominantWaveFreq(float *amplitudeArray, float sampleTime)
{
	float Amax = 0.0f;
	unsigned int w_index = 0;
	for (int i = 10; i < BUFFER_LENGTH/8; i++) // Avoid the 10 lowest frequency bins (ignores periods >~20s), and only use the lowest frequencies (<~8 rad/s for FFT sample time 0.1s)
	{
		if (*(amplitudeArray+i) > Amax)
		{
			w_index = i;
			Amax = *(amplitudeArray + i);
		}
	}

	// Find dominant frequency = > position of dominant wave in window * sampling frequency
	float f_d = w_index / (float)BUFFER_LENGTH * 1.0f / sampleTime;
	return 2.0f*PI*f_d;
}

float ParamEstimation::getDominantWaveAmplitude(float w, float* accelerationData, float sigma2)
{
	float32_t a_mean;
	unsigned int N = BUFFER_LENGTH;
	arm_mean_f32(accelerationData, N, &a_mean);
	float sum = 0.0f;	

	for (unsigned int i = 0; i < N; i++)
	{
		sum += sq(*(accelerationData + i) - a_mean);
	}
	
	float A_squared = 1.0f / (pow(w, 4))*(2.0f/N*sum-2*sigma2);
	float A;	// Amplitude of dominant wave
	if (A_squared > 0.0f)
	{
		arm_sqrt_f32(A_squared, &A);
	}
	else
	{
		A = 0.0f;
	}
	return A;
}

void ParamEstimation::getFilterParams(float w, float A, float dw, float sampleTimeAcc, float sigma2, float T_2max, float S_nn, FilterParam* params)
{	
	float w1 = w + dw;
	float w2 = w - dw;
	float k = (w2*abs(sq(w) - sq(w1)) + w1*abs(sq(w) - sq(w2))) / (2 * pow(w,3.0f));	// Schaut 2.24, the relative improvement
	
	params->w_c = pow((S_nn / (8 * sqrt(2.0f)*sq(k) * sq(A))), (1.0f / 3.0f));			// Schaut 2.26 (based on 2.12)

	params->w_c = max(params->w_c, 5.53f / T_2max);		// Schaut 2.15

	params->a0 = sqrt(2.0f)*sq(w) / params->w_c;		// Schaut equation 2.19
	params->b0 = params->a0;							// -"-
	
	params->a1 = -(sq(w) + sq(params->w_c)) / sq(w);	// Schaut equation 2.20
}

void ParamEstimation::RealFFT(float *source, float *destination) { 
	uint32_t ifftFlag = 0;		// 0 = Forward, 1 = Inverse
	uint32_t doBitReverse = 1;	// 1 = Output in normal order
	arm_rfft_instance_f32 S;
	arm_cfft_radix4_instance_f32 S_CFFT;
	float complexOutput[BUFFER_LENGTH * 2];

	// Initialize the CFFT/CIFFT module 
	arm_status status = arm_rfft_init_f32(&S, &S_CFFT, BUFFER_LENGTH, ifftFlag, doBitReverse);	// The functions support lengths of [32, 64, 128, ..., 4096] 
	if (status != ARM_MATH_SUCCESS) {
		HLSerial.Send(OutGoingCmd::errorGeneral, "rfft_init failed during parameter calculation");
		return;
	}

	// Process the data through the RFFT module
	arm_rfft_f32(&S, source, complexOutput);	// TODO: Should be changed with arm_rfft_fast_f32 or fixed point function
	
	// Process the data through the Complex Magnitude Module for calculating the magnitude at each bin 
	arm_cmplx_mag_f32(complexOutput, destination, BUFFER_LENGTH);
}
