﻿using System;
using System.Collections.Generic;
using System.IO.Ports;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;

namespace HLC.HeaveLockCommunication
{
    public class HLComm
    {
        private HeaveLockMessage HLmsg;
        private HeaveLockMessage warningMsg;
        private SerialPort _serialPort;
        private volatile bool _keepReadSerialAlive = true;
        private int _baudrate;
        private string _comPort;

        public event EventHandler<HeaveLockMessageEventArgs> serialMessageReceivedEvent;

        private enum ReadSerialState
        {
            LOOKING_4_HEADER,
            LOOKING_4_CMD,
            LOOKING_4_DATA_LENGTH,
            LOOKING_4_DATA,
            LOOKING_4_TAIL
        };

        public HLComm(int serialBaudRate, string serialPortName)
        {           
            HLmsg = new HeaveLockMessage();
            warningMsg = new HeaveLockMessage();
            _baudrate = serialBaudRate;
            _comPort = serialPortName;
            _serialPort = new SerialPort(_comPort, _baudrate);
            _serialPort.Open();
            _serialPort.DiscardInBuffer();
        }

        public bool IsAlive { get { return _keepReadSerialAlive; } }
        public bool SerialPortIsOpen { get { return _serialPort.IsOpen; } }

        private void OnRaiseMessageReceivedEvent(HeaveLockMessageEventArgs e)
        {
            // Make a temporary copy - avoid racecondition
            EventHandler<HeaveLockMessageEventArgs> handler = serialMessageReceivedEvent;
            // Event will be null if there are no subscribers
            if (handler != null)
            {
                handler(this, e);
            }
        }
        /// <summary>
        /// ReadAndExtractSerial strips the header and tail byte from a message from the HeaveLock.
        /// It checks that the message-command used is valid and then publishes the valid data to an event.
        /// This method is ment to be run as a seperate thread.
        /// </summary>
        public void ReadSerial()
        {            
            while (_keepReadSerialAlive)
            {                
                ReadAndExtractSerial();
                Thread.Sleep(1); //allow other stuff to happen
            }
            try {
                _serialPort.DiscardOutBuffer();
                _serialPort.DiscardInBuffer();
                _serialPort.Close();
            }
            catch (Exception) { }
        }

        /// <summary>
        /// Terminates ReadSerial()
        /// </summary>
        public void stopRead()
        {
            _keepReadSerialAlive = false;
        }
        public void setBaudRate(int baudrate)
        {
            _baudrate = baudrate;
            _serialPort.BaudRate = baudrate;
        }

        //tips and tricks: https://blogs.msdn.microsoft.com/bclteam/2006/10/10/top-5-serialport-tips-kim-hamilton/
        private void ReadAndExtractSerial()
        {
            try
            {
                if (_serialPort.IsOpen && _serialPort.BytesToRead > 0)
                {
                    ReadSerialState state = ReadSerialState.LOOKING_4_HEADER;
                    Stopwatch _now = new Stopwatch();
                    byte b;
                    List<byte> buffer = new List<byte>();
                    HLmsg.cmd = 0x00;
                    byte numBytesToRead = 0;
                    byte dataCount = 0;
                    _now.Start();
                    while (_now.ElapsedMilliseconds < 15)
                    {
                        if (_serialPort.BytesToRead > 0)
                            b = (byte)_serialPort.ReadByte();
                        else
                            continue;

                        switch (state)
                        {
                            case ReadSerialState.LOOKING_4_HEADER:
                                if (b == IncomingCmd.header)
                                    state = ReadSerialState.LOOKING_4_CMD;
                                break;
                            case ReadSerialState.LOOKING_4_CMD:
                                if (IncomingCmd.ValidateCmd(b))
                                {
                                    HLmsg.cmd = b;
                                    state = ReadSerialState.LOOKING_4_DATA_LENGTH;
                                }
                                break;
                            case ReadSerialState.LOOKING_4_DATA_LENGTH:
                                numBytesToRead = b;
                                if (numBytesToRead > 0)
                                    state = ReadSerialState.LOOKING_4_DATA;
                                else
                                    state = ReadSerialState.LOOKING_4_TAIL;
                                break;
                            case ReadSerialState.LOOKING_4_DATA:
                                buffer.Add(b);
                                dataCount++;
                                if (dataCount >= numBytesToRead)
                                    state = ReadSerialState.LOOKING_4_TAIL;
                                break;
                            case ReadSerialState.LOOKING_4_TAIL:
                                if (b != IncomingCmd.tail)
                                {
                                    OnRaiseMessageReceivedEvent(new HeaveLockMessageEventArgs(CreateWarning("HLC: Tail not found when expected. Command used in missed message: ", HLmsg.cmd)));
                                    return; //error
                                }
                                else
                                {
                                    HLmsg.data = buffer.ToArray();
                                    OnRaiseMessageReceivedEvent(new HeaveLockMessageEventArgs(HLmsg)); //publish new message as event
                                    return; //done
                                }
                        }
                    }
                    OnRaiseMessageReceivedEvent(new HeaveLockMessageEventArgs(CreateWarning("HLC: Serialread timed out locally. Command used in missed message: ", HLmsg.cmd))); //serial timed out
                }
            }
            catch (System.InvalidOperationException)
            {
                //fires if serial disconnected suddenly (i.e. USB disconnected) 
                //do nothing, falls through. If it's disconnected it should not go into this loop again
            }
            catch (System.IO.IOException)
            {
                //fires if serial disconnected suddenly (i.e. USB disconnected) 
                //do nothing, falls through. If it's disconnected it should not go into this loop again
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error trying to read serial: " + ex);
            }

        }

        private HeaveLockMessage CreateWarning(string warning, byte cmd)
        {
            warning += cmd.ToString("X");
            if(cmd == 0)
            {
                warning += ". Baudrate may be wrong. Change baud rate and reconnect."; 
            }
            List<byte> errorBuf = new List<byte>();
            foreach (byte c in warning)
            {
                errorBuf.Add(c);
            }
            warningMsg.cmd = IncomingCmd.errorGeneral;
            warningMsg.data = errorBuf.ToArray();
            return warningMsg;
        }
        
        public void SerialSend(byte cmd)
        {
            List<byte> wrappedMsg = new List<byte>();
            wrappedMsg.Add(OutGoingCmd.header);
            wrappedMsg.Add(cmd);
            wrappedMsg.Add(0x00); //no length
            wrappedMsg.Add(OutGoingCmd.tail);
            WriteToSerial(wrappedMsg.ToArray());
        }

        public void SerialSend(byte cmd, float value)
        {
            List<byte> wrappedMsg = new List<byte>();

            wrappedMsg.Add(OutGoingCmd.header);
            wrappedMsg.Add(cmd);
            wrappedMsg.Add((byte)BitConverter.GetBytes(value).GetLength(0)); // 0 represent the first dimension of the array, in this case there is only one dimension
            wrappedMsg.AddRange(BitConverter.GetBytes(value));
            wrappedMsg.Add(OutGoingCmd.tail);
            WriteToSerial(wrappedMsg.ToArray());
        }

        public void SerialSend(byte cmd, uint value)
        {
            List<byte> wrappedMsg = new List<byte>();
            wrappedMsg.Add(OutGoingCmd.header);
            wrappedMsg.Add(cmd);
            wrappedMsg.Add((byte)BitConverter.GetBytes(value).GetLength(0)); // 0 represent the first dimension of the array, in this case there is only one dimension
            wrappedMsg.AddRange(BitConverter.GetBytes(value));
            wrappedMsg.Add(OutGoingCmd.tail);
            WriteToSerial(wrappedMsg.ToArray());
        }

        public void SerialSend(byte cmd, float[] values)
        {
            int numberOfValues = values.Length;
            List<byte> wrappedMsg = new List<byte>();
            wrappedMsg.Add(OutGoingCmd.header);
            wrappedMsg.Add(cmd);
            wrappedMsg.Add((byte)(values.Length * 4));
            for(int i = 0; i < values.Length; i++)
            {
                wrappedMsg.AddRange(BitConverter.GetBytes(values[i]));
            }
            wrappedMsg.Add(OutGoingCmd.tail);
            WriteToSerial(wrappedMsg.ToArray());
        }
        
        private void WriteToSerial(byte[] message)
        {
            try
            {
                _serialPort.Write(message, 0, message.Length);
            }
            catch (System.InvalidOperationException) {/*do nothing, falls through*/}
            catch(System.IO.IOException) {/*do nothing, falls through*/}
            catch (Exception ex)
            {
                MessageBox.Show("Write to serial failed: " + ex);
            }
        }
    }

  
}
